# -*- coding: utf-8 -*-
"""
Created on Tue May 12 22:20:36 2020

@author: yoelr
"""

__all__ = ('Predictor',)

class Predictor:
    """Abstract Predictor class."""
    
    def __init_subclass__(cls):
        assert hasattr(cls, 'fit'), 'Predictor must implement a `fit` method'
        assert hasattr(cls, 'predict'), 'Predictor must implement a `predict` method'
        assert hasattr(cls, 'get_params'), 'Predictor must implement a `get_params` method'
        
    def __init__(self, metric):
        self.metric = metric
    
    def __repr__(self):
        return f"<{type(self).__metric__}: {self.metric}>"


# -*- coding: utf-8 -*-
"""
Created on Tue May 12 23:02:00 2020

@author: yoelr
"""
__all__ = ('Normalizer',)

class Normalizer:
    __slots__ = ('width', 'minimum')
    
    def __init__(self, width, minimum):
        self.width = width
        self.minimum = minimum

    @classmethod
    def from_data(cls, X, axis=None):
        if axis is None:
            ndim = X.ndim
            if ndim > 1:
                axis = list(range(ndim))
                axis.remove(axis[-1])
                axis = tuple(axis)
        maximum = X.max(axis=axis, keepdims=True)
        minimum = X.min(axis=axis, keepdims=True)
        width = maximum - minimum
        return cls(width, minimum)
    
    def __call__(self, X, scale=True):
        return self.scale(X) if scale else self.unscale(X)
    
    def scale(self, X):
        return (X - self.minimum) / self.width
    
    def unscale(self, X):
        return (X * self.width) + self.minimum
    
    def __repr__(self):
        return f"<{type(self).__name__}>"
    


# -*- coding: utf-8 -*-
"""
Created on Tue May 12 18:37:59 2020

@author: yoelr
"""
import numpy as np
from ._rescaler import Rescaler
from ._normalizer import Normalizer
from ._predictor import Predictor
from ._multi_predictor import MultiPredictor


__all__ = ('FittedModel',)

def assert_within_bounds(name, lb, xi, ub):
    assert (lb < xi < ub).all(), f"{name} ({xi:.5g}) is not within bounds {(lb, ub)}"

class FittedModel:
    __slots__ = ('predictor', 'preprocessor', 'postprocessor', 'bounds', 'features')
    
    def __init__(self, predictor, preprocessor, postprocessor, bounds=None, features=None):
        self.predictor = predictor
        self.preprocessor = preprocessor
        self.postprocessor = postprocessor
        self.bounds = bounds
        self.features = features
        
    def predict(self, X, check_bounds=True):
        X = np.asarray(X, float)
        if check_bounds: self.check_bounds(X)
        X = self.preprocessor(X)
        y = self.predictor.predict(X)
        return self.postprocessor(y, False)
    
    @property
    def metrics(self):
        predictor = self.predictor
        if isinstance(predictor, Predictor):
            return [predictor.metric]
        elif isinstance(predictor, MultiPredictor):
            return predictor.metrics
        else:
            raise AttributeError('could not find metric names')
    
    @classmethod
    def from_dfs(cls, Xdf, ydf, ML=None, bounds=None, features=None):
        X = Xdf.values
        y = ydf.values
        preprocessor = Rescaler.from_data(X)
        postprocessor = Normalizer.from_data(y)
        if not ML: from sklearn.svm import SVR as ML
        if ydf.ndim == 1:
            predictor = ML()
            predictor.metric = ydf.name
        else:
            def create_predictor(metric):
                ml = ML()
                ml.metric = metric
                return ml
            predictors = [create_predictor(col) for col in ydf]
            predictor = MultiPredictor(predictors)
        if not features: features = tuple(Xdf)
        self = cls(predictor, preprocessor, postprocessor, bounds, features)
        self.fit(X, y)
        return self
    
    def fit(self, X, y):
        X = self.preprocessor(X)
        y = self.postprocessor(y)
        self.predictor.fit(X, y)
    
    def check_bounds(self, xs):
        bounds = self.bounds
        if bounds is None: return
        names = self.features
        if names:
            for name, xi, (lb, ub) in zip(names, xs.transpose(), bounds):
                assert_within_bounds(name, lb, xi, ub)
        else:
            for i, (xi, (lb, ub)) in enumerate(zip(xs.transpose(), bounds)):
                name = f"Feature #{i}"
                assert_within_bounds(name, lb, xi, ub)
            
    def __repr__(self):
        newline = "\n" + " " * len(" metrics: ")
        metric_info = newline.join([str(i) for i in self.metrics])
        features = self.features
        if features:
            newline = "\n" + " " * len(" features: ")
            feature_info = newline.join([str(i) for i in features])
        else:
            feature_info = "?"
        return (f"{type(self).__name__}:\n"
                f" metrics: {metric_info}\n"
                f" features: {feature_info}")


# -*- coding: utf-8 -*-
"""
Created on Tue May 12 21:52:28 2020

@author: yoelr
"""
import numpy as np

__all__ = ('MultiPredictor',)

class MultiPredictor:
    __slots__ = ('predictors',)
    
    def __init__(self, predictors):
        self.predictors = predictors
        
    def fit(self, X, y):
        for predictor, yi in zip(self.predictors, y.transpose()): predictor.fit(X, yi)
        
    def predict(self, xs):
        return np.array([i.predict(xs) for i in self.predictors], float).transpose()
    
    @property
    def metrics(self):
        return [i.metric for i in self.predictors]
    
    def __repr__(self):
        names = ', '.join([str(i) for i in self.metrics])
        return f"<{type(self).__name__}: {names}>"
        


# -*- coding: utf-8 -*-
"""
Created on Tue May 12 22:02:45 2020

@author: yoelr
"""
from . import _rescaler
from . import _normalizer
from . import _predictor
from . import _multi_predictor
from . import _fitted_model

__all__ = (*_rescaler.__all__,
           *_normalizer.__all__,
           *_multi_predictor.__all__,
           *_fitted_model.__all__,
           *_predictor.__all__,
)

from ._rescaler import *
from ._normalizer import *
from ._predictor import *
from ._multi_predictor import *
from ._fitted_model import *



# -*- coding: utf-8 -*-
"""
Created on Sat Nov 18 16:17:00 2017

@author: Yoel Cortes-Pena
"""
from setuptools import setup
#from Cython.Build import cythonize
#import numpy

setup(
    name='pipeml',
    packages=['pipeml'],
    license='MIT',
    version='0.1',
    description='Pipeline tools for machine learning models',
    long_description=open('README.rst').read(),
    author='Yoel Cortes-Pena',
    install_requires=['sklearn', 'numpy'],
    python_requires=">=3.6",
    package_data=
        {'pipeml': []},
    platforms=['Windows', 'Mac', 'Linux'],
    author_email='yoelcortes@gmail.com',
    url='https://github.com/yoelcortes/pipeml',
    download_url='https://github.com/yoelcortes/pipeml.git',
    classifiers=['Development Status :: 3 - Alpha',
                 'Environment :: Console',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python :: 3.6',
				 'Programming Language :: Python :: 3.7',
                 'Topic :: Scientific/Engineering'],
    keywords='machine learning tools function pipeline',
)


# -*- coding: utf-8 -*-
"""
Created on Mon Nov 11 07:38:24 2019

@author: yoelr
"""
__all__ = ('Rescaler',)

class Rescaler:
    __slots__ = ('mean', 'std', 'factor')
    
    @classmethod
    def from_data(cls, X, factor=0.5, axis=None):
        if axis is None:
            axis = list(range(X.ndim))
            axis.remove(axis[-1])
            axis = tuple(axis)
        return cls(X.mean(axis=axis, keepdims=True),
                   X.std(axis=axis, keepdims=True),
                   factor)
    
    @staticmethod
    def rescale(X, factor=0.5, dim=-1):
        axis = list(range(X.ndim))
        axis.remove(axis[dim])
        axis = tuple(axis)
        mean = X.mean(axis=axis, keepdims=True),
        std = X.std(axis=axis, keepdims=True),
        return factor*(X - mean) / std
    
    def __init__(self, mean, std, factor):
        self.mean = mean
        self.std = std
        self.factor = factor
        
    def __call__(self, X, scale=True):
        return self.scale(X) if scale else self.unscale(X)
        
    def scale(self, X):
        return self.factor*(X - self.mean) / self.std
    
    def unscale(self, X):
        return (X * self.std / self.factor) + self.mean
    
    def __repr__(self):
        return f"<{type(self).__name__}(factor={self.factor})>"